
import pytest
from sanic import response, Sanic

from sanicargs import parse_query_args, fields
from sanic.websocket import WebSocketProtocol
from sanic.exceptions import InvalidUsage
from sanic import request

import datetime

from functools import wraps
import inspect

def has_test_arg(func):
    signature = inspect.signature(func)
    assert signature.parameters['test']
    @wraps(func)
    async def decorated(request, *args, **kwargs):
        return await func(request, *args, **kwargs)
    return decorated


@pytest.yield_fixture
def app():
    app = Sanic("test_sanic_app")

    @app.route("/int", methods=['GET'])
    @parse_query_args
    async def test_int(request, test: int):
        return response.json({'test': test})

    @app.route("/str", methods=['GET'])
    @parse_query_args
    async def test_str(request, test: str):
        return response.json({'test': test})

    @app.route("/bool", methods=['GET'])
    @parse_query_args
    async def test_bool(request, test: bool):
        return response.json({'test': test})

    @app.route("/datetime", methods=['GET'])
    @parse_query_args
    async def test_datetime(request, test: datetime.datetime):
        return response.json({'test': test.isoformat()})

    @app.route("/date", methods=['GET'])
    @parse_query_args
    async def test_date(request, test: datetime.date):
        return response.json({'test': test.isoformat()})

    @app.route("/list", methods=['GET'])
    @parse_query_args
    async def test_list(req, test: fields.List[str] = None):
        return response.json({'test': test})

    @app.route("/all", methods=['GET'])
    @parse_query_args
    async def test_all(
            req: request, 
            a: int,
            b: str,
            c: datetime.datetime,
            d: datetime.date,
            e: fields.List[str] = None):
        return response.json({
            'a': a,
            'b': b,
            'c': c.isoformat(),
            'd': c.isoformat(),
            'e': e
        })

    @app.route("/optional", methods=['GET'])
    @parse_query_args
    async def test_optional(request, test: str = 'helloworld'):
        return response.json({'test': test})

    @app.route("/with/<path_param>/path_params", methods=['GET'])
    @parse_query_args
    async def test_path_params(request, path_param: int, test: str, test_2: int=35):
        return response.json({'path_param': path_param, 'test': test, 'test_2': test_2})

    @app.route("/test_arg", methods=['GET'])
    @has_test_arg
    @parse_query_args
    async def test_args(request, test: int):
        return response.json({'test': test})

    yield app

@pytest.fixture
def test_cli(loop, app, test_client):
    return loop.run_until_complete(test_client(app, protocol=WebSocketProtocol))


#########
# Tests #
#########


async def test_parse_int_success(test_cli):
    resp = await test_cli.get('/int?test=10')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': 10}


async def test_parse_int_fail(test_cli):
    resp = await test_cli.get('/int?test=not an integer')
    assert resp.status == 400


async def test_parse_bool_true_success(test_cli):
    resp = await test_cli.get('/bool?test=true')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': True }

async def test_parse_bool_false_success(test_cli):
    resp = await test_cli.get('/bool?test=false')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': False }

async def test_parse_bool_fail(test_cli):
    resp = await test_cli.get('/bool?test=not an bool')
    assert resp.status == 400


async def test_parse_str_success(test_cli):
    resp = await test_cli.get('/str?test=hello')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': 'hello'}


async def test_parse_str_also_works_with_int(test_cli):
    ''' there is no way of knowing if an str is really an integer
    '''
    resp = await test_cli.get('/str?test=400')
    assert resp.status == 200


async def test_parse_datetime_success(test_cli):
    resp = await test_cli.get('/datetime?test=2017-10-10T10:10:10')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': '2017-10-10T10:10:10'}


async def test_parse_datetime_fail(test_cli):
    resp = await test_cli.get('/datetime?test=not a datetime')
    assert resp.status == 400


async def test_parse_date_success(test_cli):
    resp = await test_cli.get('/date?test=2017-10-10')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': '2017-10-10'}


async def test_parse_date_fail(test_cli):
    resp = await test_cli.get('/date?test=not a datetime')
    assert resp.status == 400
   

async def test_parse_list_success(test_cli):
    resp = await test_cli.get('/list?test=one,two,three')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': [
      'one', 'two', 'three'
    ]}


async def test_parse_list_also_works_with_singular(test_cli):
    resp = await test_cli.get('/list?test=not a datetime')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': [
      'not a datetime'
    ]}


async def test_all_at_once(test_cli):
    resp = await test_cli.get('/all?a=10&b=test&c=2017-10-10T10:10:10&d=2017-10-10&e=a,b,c,d,e')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == dict(
      a=10,
      b='test',
      c='2017-10-10T10:10:10',
      d='2017-10-10T10:10:10',
      e=[
        'a', 'b', 'c', 'd', 'e'
      ]
    )


async def test_optional(test_cli):
    resp = await test_cli.get('/optional')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': 'helloworld'}


async def test_mandatory(test_cli):
    resp = await test_cli.get('/str')
    assert resp.status == 400


async def test_with_path_params(test_cli):
    resp = await test_cli.get('/with/123/path_params?test=hello')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'path_param': 123, 'test': 'hello', 'test_2': 35}

async def test_args_success(test_cli):
    resp = await test_cli.get('/test_arg?test=10')
    assert resp.status == 200
    resp_json = await resp.json()
    assert resp_json == {'test': 10}



import inspect
import datetime
from functools import wraps

from sanic import response
from sanic.exceptions import abort

from sanicargs.fields import List

from logging import getLogger

__logger = getLogger('sanicargs')


def __parse_datetime(str):
    # attempt full date time, but tolerate just a date
    try:
        return datetime.datetime.strptime(str, '%Y-%m-%dT%H:%M:%S')
    except Exception:
        pass
    return datetime.datetime.strptime(str, '%Y-%m-%d')


def __parse_date(str):
    return datetime.datetime.strptime(str, '%Y-%m-%d').date()


def __parse_bool(str):
    lower = str.lower()
    if lower == 'true':
        return True
    if lower == 'false':
        return False
    raise ValueError("Can't parse {} as boolean".format(str))


__type_deserializers = {
    bool: __parse_bool,
    int: int,
    str: str,
    datetime.datetime: __parse_datetime,
    datetime.date: __parse_date,
    List[str]: lambda s: s.split(',')
}


def parse_query_args(func):
    '''parses query args and validates, deserializes them
    VERY IMPORTANT!:
    to use this decorator it must be used in a Sanic endpoint and used BEFORE the 
    sanic blueprint decorator like so:
        @blueprint.route("/foo/<businessunitid>/bar")
        @authorize_business_unit
        @parse_query_args
    and the signature of the function needs to start with request and the rest of 
    the parameters need type hints like so:
        async def generate_csv(request, query: str, businessunitid: str):
    '''
    notations = inspect.signature(func)

    parameters = [
        (name, p.annotation, p.default)
        for name, p in notations.parameters.items()
    ]
    request_arg_name = inspect.getfullargspec(func)[0][0]
    
    @wraps(func)
    async def inner(request, *old_args, **route_parameters):
        kwargs = {}
        name = None
        try:
            for name, arg_type, default in parameters:
                raw_value = request.args.get(name, None)

                # provided in route
                if name in route_parameters or name==request_arg_name:
                    if name==request_arg_name:
                        continue
                    raw_value = route_parameters[name]

                # no value
                elif name not in request.args:
                    if default != inspect._empty:
                        # TODO clone?
                        kwargs[name] = default
                        continue
                    else:
                        raise KeyError("Missing required argument %s" % name)

                parsed_value = __type_deserializers[arg_type](raw_value)
                kwargs[name] = parsed_value
        except Exception as err: 
            __logger.warning({
                "message": "Request args not validated",
                "stacktrace": str(err)
            })
            return abort(400, 'Bad or missing value for %s' % name)
        return await func(request, **kwargs)
    return inner



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import os.path

try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

readme = read_md('README.md')

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'sanic>=0.7.0'
]

test_requirements = [
    'pytest==3.2.0',
    'pytest-sanic>=0.1.6'
]

extras = {
    'test': test_requirements + requirements
}

# get version
metadata = {}
version_filename = os.path.join(os.path.dirname(__file__), 'sanicargs','__version__.py')
exec(open(version_filename).read(), None, metadata)

setup(
    name='sanicargs',
    version=metadata['__version__'],
    description="Parses query args in sanic using type annotations",
    long_description=readme + '\n\n' + history,
    author="jgv",
    author_email='jgv@trustpilot.com',
    url='https://github.com/trustpilot/python-sanicargs',
    packages=[
        'sanicargs',
    ],
    package_dir={'sanicargs':
                 'sanicargs'},
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords='sanicargs sanic query args type annotations',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    extras_require=extras,

)
